#!/usr/bin/env python2.7
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Create simple self-extracting archives."""

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

__metaclass__ = type

import argparse
import base64
from os import (
    stat,
    walk,
    )
from os.path import (
    curdir,
    dirname,
    isfile,
    join,
    normpath,
    relpath,
    )
from pipes import quote
from sys import (
    stderr,
    stdout,
    )


class ShellArchive:
    """Create a simple shell archive.

    Relies on ``coreutils`` and ``/bin/sh`` (e.g. ``dash``).
    """

    def __init__(self):
        super(ShellArchive, self).__init__()
        self.members = {}

    def add(self, inpath, outpath):
        """Add a file to the archive.

        :param inpath: The filename of the file to copy in.
        :param outpath: The filename to extract to.
        """
        outpath = normpath(outpath)
        self.members[outpath] = inpath

    def scan(self, root):
        """Recursively add a directory to the archive."""
        for dirpath, dirnames, filenames in walk(root):
            for filename in filenames:
                inpath = join(dirpath, filename)
                if isfile(inpath):
                    outpath = relpath(join(dirpath, filename), root)
                    self.add(inpath, outpath)

    def write(self, fout, execute=None):
        write = fout.write
        write(b"#!/bin/sh\n")
        write(b"set -o errexit\n")
        write(b"set -o noclobber\n")
        outdirs = {b""}  # Already seen output directories.
        for outfile, infile in sorted(self.members.items()):
            write(b"\n")
            # Create the output directory if we haven't already.
            outdir = dirname(outfile)
            if outdir not in outdirs:
                write(b"mkdir --parents -- %s\n" % quote(outdir))
                outdirs.add(outdir)
                write(b"\n")
            # Echo the name of the file about to be extracted.
            write(b"printf '%%s\\n' %s\n" % quote(outfile))
            # Base64 decode the heredoc into the output file
            write(b"base64 --decode << '~~~' > %s\n" % quote(outfile))
            # Base64 encode the input file into the archive.
            with open(infile, "rb") as fin:
                base64.encode(fin, fout)
            # End the heredoc; tilde is not a valid base64 character.
            write(b"~~~\n")
            # Copy over the user execute bit.
            if stat(infile).st_mode & 0100 > 0:
                write(b"chmod u+x -- %s\n" % quote(outfile))
        # Replace the shell by executing one of the archive members.
        if execute is not None:
            execute = normpath(execute)
            if execute in self.members:
                execute = join(curdir, execute)
                write(b"\nexec %s\n" % quote(execute))
            else:
                raise LookupError(
                    "%r is not an archive member" % execute)


# See http://docs.python.org/release/2.7/library/argparse.html.
argument_parser = argparse.ArgumentParser(description=__doc__)
argument_parser.add_argument(
    "source", metavar="DIR", help=(
        "All regular files in this directory and all subdirectories "
        "will be added to the archive. File paths in the archive will "
        "be recorded relative to this directory."))
argument_parser.add_argument(
    "--execute", metavar="EXECUTE", default=None, help=(
        "The path of a script or other executable to exec upon successful "
        "extraction of the archive. This path should be relative to the "
        "source directory, i.e. the path of the file as extracted."
        ))


if __name__ == "__main__":
    try:
        args = argument_parser.parse_args()
        archive = ShellArchive()
        archive.scan(args.source)
        archive.write(stdout, args.execute)
    except KeyboardInterrupt:
        raise SystemExit(1)
    except StandardError as error:
        print(error, file=stderr)
        raise SystemExit(2)
    else:
        raise SystemExit(0)
